from importlib import import_module
import ast
import copy
import sys

import yaml

from .utils import deep_merge, set_by_path


# These default values are automatically set if not specified explicitly.
_defaults = {'random_seed': 12345}

# These keys are required and have to be set within the configuration.
_required = []


class TrainingConfig:

    def __init__(self, config_files):
        self._config = {}

        # read in all config files
        for file in config_files:
            self._add_config(file)

        self._parse_cmd_args()
        self._set_defaults()
        self._raw_config = copy.deepcopy(self._config)
        self._instantiate_components()

        self._check()

    def get_raw_config(self):
        return self._raw_config

    def _add_config(self, path):
        with open(path) as fh:
            yaml_dict = yaml.load(fh.read(), Loader=yaml.FullLoader)
        self._config = deep_merge(self._config, yaml_dict)

    def _parse_cmd_args(self):
        try:
            i = sys.argv.index('with')
        except ValueError:
            # no config options passed
            return

        config_options = sys.argv[(i + 1):]
        if len(config_options) == 0:
            return

        # parse each config option individually
        for option in config_options:
            before, match, after = option.partition('=')
            if match == '=':
                try:
                    value = ast.literal_eval(after)
                except (ValueError, SyntaxError):
                    value = after  # interpret as string
                set_by_path(self._config, before, value)
            else:
                # named config
                raise NotImplementedError

    def _set_defaults(self):
        for key, value in _defaults.items():
            if key not in self._config:
                self._config[key] = value

    def _instantiate_components(self):
        def recursive_init(dic, key):
            value = dic[key]
            if isinstance(value, dict) and '__type__' in value:
                assert isinstance(value['__type__'], str), '__type__ has to be of type "str".'

                parts = value['__type__'].split('.')
                assert len(parts) > 1, 'Unrecognized type definition'

                module_path = '.'.join(parts[:-1])
                type_name = parts[-1]
                _type = getattr(import_module(module_path), type_name)

                # initialize
                del value['__type__']
                dic[key] = _type(**value)
            elif isinstance(value, dict):
                for key in value.keys():
                    recursive_init(value, key)

        for key in self._config:
            recursive_init(self._config, key)

    def _check(self):
        missing = set(_required).difference(self._config.keys())
        if missing:
            raise ValueError(f'Training configuration is incomplete. Missing {missing}')

    def __getitem__(self, item):
        return self._config[item]

    def items(self):
        return self._config.items()

    def __repr__(self):
        return f'{__class__.__name__}(' + ', '.join([f'{k}={repr(v)}' for k, v in self._config.items()]) + ')'



import inspect
from abc import ABC, abstractmethod
from pathlib import Path
from typing import List

import yaml
from torch import nn

from .bound_functions import BoundFunctions


# TODO implement standardlogger and always use it


class BaseLogger(ABC):

    def __init__(self):
        pass

    @abstractmethod
    def setup(self, save_path: Path, run_name: str):
        """
        Used to setup the logger, given the run name
        """
        pass

    @abstractmethod
    def before_training_start(self, config: dict, model: nn.Module, bound_functions: BoundFunctions):
        pass

    @abstractmethod
    def after_pass(self, metrics: dict, epoch: int, is_validate: bool = False):
        pass

    @abstractmethod
    def after_epoch(self, epoch: int, model: nn.Module):
        pass


class StandardLogger(BaseLogger):

    def __init__(self):
        super().__init__()
        self._save_path = None
        self._log_file = None

    def setup(self, save_path: Path, run_name: str):
        self._save_path = save_path
        self._log_file = self._save_path / 'log.txt'

    def before_training_start(self, config: dict, model: nn.Module, bound_functions: BoundFunctions):
        # save config to file
        with (self._save_path / 'config.yaml').open('w') as fh:
            yaml.dump(config, fh)

        # save model summary and source to file
        with (self._save_path / 'model_summary.txt').open('w') as fh:
            fh.write(repr(model))

        with (self._save_path / 'source.txt').open('a') as fh:
            fh.write(inspect.getsource(type(model)))
            for function in bound_functions.values():
                fh.write('\n')
                fh.write(inspect.getsource(function))

    def after_pass(self, metrics: dict, epoch: int, is_validate: bool = False):
        status_str = f'[Epoch {epoch} / {"Val" if is_validate else "Train"}] ' \
                     + ' '.join([f'{k}: {v}' for k, v in metrics.items()]) + '\n'
        with self._log_file.open('a') as fh:
            fh.write(status_str)

    def after_epoch(self, epoch: int, model: nn.Module):
        pass


class WandbLogger(BaseLogger):

    def __init__(self, **kwargs):
        super().__init__()
        import wandb
        self.wandb = wandb
        self._wandb_args = kwargs

    def setup(self, save_path: Path, run_name: str):
        if 'name' in self._wandb_args:
            del self._wandb_args['name']
        self.wandb.init(name=run_name, dir=save_path, **self._wandb_args)

    def before_training_start(self, config: dict, model: nn.Module, bound_functions: BoundFunctions):
        self.wandb.config.update(config)
        # self.wandb.watch(model)

    def after_pass(self, metrics: dict, epoch: int, is_validate: bool = False):
        prefix = 'val_' if is_validate else 'train_'
        self.wandb.log({(prefix + k): v for k, v in metrics.items()}, step=epoch)

    def after_epoch(self, epoch: int, model: nn.Module):
        pass


class LoggerSet(BaseLogger):

    def __init__(self, loggers: List[BaseLogger]):
        super().__init__()
        self._loggers = loggers

    def setup(self, save_path: Path, run_name: str):
        for logger in self._loggers:
            logger.setup(save_path, run_name)

    def before_training_start(self, config: dict, model: nn.Module, bound_functions: BoundFunctions):
        for logger in self._loggers:
            logger.before_training_start(config, model, bound_functions)

    def after_pass(self, metrics: dict, epoch: int, is_validate: bool = False):
        for logger in self._loggers:
            logger.after_pass(metrics, epoch, is_validate)

    def after_epoch(self, epoch: int, model: nn.Module):
        for logger in self._loggers:
            logger.after_epoch(epoch, model)
